#pragma once
#include <vector>
#include <fishnet/Shapefile.hpp>
#include <fishnet/ShapeGeometry.hpp>
#include <fishnet/VectorLayer.hpp>
#include <fishnet/PolygonNeighbours.hpp>
#include <fishnet/GraphFactory.hpp>
#include <fishnet/CompositePredicate.hpp>
#include <fishnet/Rectangle.hpp>
#include <fishnet/WGS84Ellipsoid.hpp>
#include "MemgraphAdjacency.hpp"
#include "SettlementPolygon.hpp"
#include "Task.hpp"
#include "FindNeighboursConfig.hpp"

/**
 * @brief Implementation of the neighbours task. Finds neighbouring settlements according to the neighbouring predicates specified in the configuration file
 * 
 * @tparam P polygon type
 */
template<fishnet::geometry::IPolygon P>
class FindNeighboursTask : public Task{
private:
    FindNeighboursConfig config;
    fishnet::Shapefile primaryInput;
    std::vector<fishnet::Shapefile> additionalInput;
    DistanceFunction distanceFunction;

    using number = typename P::numeric_type;
    struct PrimaryInputAABB{
        number left = std::numeric_limits<number>::max();
        number right = std::numeric_limits<number>::min();
        number top = std::numeric_limits<number>::min();
        number bottom = std::numeric_limits<number>::max();

        void update(const P & polygon) {
            auto aaBB = fishnet::geometry::Rectangle<number>(polygon);
            if(aaBB.left() < left)
                left = aaBB.left();
            if(aaBB.right() > right)
                right = aaBB.right();
            if(aaBB.top() > top)
                top = aaBB.top();
            if(aaBB.bottom() < bottom)
                bottom = aaBB.bottom();
        }

        fishnet::geometry::Rectangle<number> asShape() const noexcept {
            return fishnet::geometry::Rectangle<number>(left,top,right,bottom);
        }
    };

public:
    FindNeighboursTask(FindNeighboursConfig && config,fishnet::Shapefile primaryInput, size_t workflowID):Task(workflowID),config(std::move(config)),primaryInput(std::move(primaryInput)){
        this->desc["type"]="NEIGHBOURS";
        this->desc["config"]=this->config.jsonDescription;
        this->desc["primary-input"] = this->primaryInput.getPath().filename().string();
    }

    FindNeighboursTask<P> & addShapefile(fishnet::Shapefile && shpFile) noexcept {
        additionalInput.push_back(std::move(shpFile));
        return *this;
    }

    std::vector<SettlementPolygon<P>> readInput(auto const & graph)  {
        std::vector<SettlementPolygon<P>> polygons;
        auto layer = fishnet::VectorLayer<P>::read(primaryInput); // load polygons from primary shapefile
        distanceFunction = distanceFunctionForSpatialReference(layer.getSpatialReference());
        if(layer.isEmpty()){
            return polygons;
        }
        PrimaryInputAABB inputBoundingBox;
        auto primaryFileRef = graph.getAdjacencyContainer().getDatabaseConnection().addFileReference(primaryInput.getPath()); // load file reference from database
        if(not primaryFileRef){
            throw std::runtime_error("Could not create file reference for shp file:\n"+primaryInput.getPath().string());
        }
        auto primaryOptFishnetIdField = layer.getSizeField(Task::FISHNET_ID_FIELD);
        if(not primaryOptFishnetIdField) {
            throw std::runtime_error("Could not find FISHNET_ID field in shp file: \n"+primaryInput.getPath().string());
        }
        for(const auto & feature : layer.getFeatures()) {
            auto optId = feature.getAttribute(primaryOptFishnetIdField.value()); // read FISHNET_ID of feature
            if(not optId){
                throw std::runtime_error("No id exists for feature with geometry:\n"+feature.getGeometry().toString());
            }
            inputBoundingBox.update(feature.getGeometry());
            polygons.emplace_back(optId.value(),primaryFileRef.value(),std::move(feature.getGeometry())); // create settlement wrapper containing its unique id and geometry
        }
        DistanceBiPredicate distanceToPrimaryInput {distanceFunction,config.maxEdgeDistance};
        fishnet::geometry::Rectangle<number> primaryInputAABB = inputBoundingBox.asShape();
        std::vector<std::string> additionalInputStrings;
        std::ranges::for_each(additionalInput,[&additionalInputStrings](auto const & file){additionalInputStrings.push_back(file.getPath().filename().string());});
        this->desc["additional-inputs"]=additionalInputStrings;
        for(const auto & shp : additionalInput) {
            auto neighbourLayer = fishnet::VectorLayer<P>::read(shp); // load polygons from shapefile
            if(not layer.getSpatialReference().IsSame(&neighbourLayer.getSpatialReference()))
                throw std::runtime_error("Spatial reference of neighbouring file does not match!\nExpecting: "+std::string(layer.getSpatialReference().GetName())+"\nActual: "+neighbourLayer.getSpatialReference().GetName());
            if(neighbourLayer.isEmpty())
                continue;
            auto fileRef = graph.getAdjacencyContainer().getDatabaseConnection().addFileReference(shp.getPath()); // load file reference from database
            if(not fileRef){
                throw std::runtime_error("Could not create file reference for shp file:\n"+shp.getPath().string());
            }
            auto optFishnetIdField = neighbourLayer.getSizeField(Task::FISHNET_ID_FIELD);
            if(not optFishnetIdField) {
                throw std::runtime_error("Could not find FISHNET_ID field in shp file: \n"+shp.getPath().string());
            }
            for(const auto & feature : neighbourLayer.getFeatures()) {
                auto optId = feature.getAttribute(optFishnetIdField.value()); // read FISHNET_ID of feature
                if(not optId){
                    throw std::runtime_error("No id exists for feature with geometry:\n"+feature.getGeometry().toString());
                }
                if(distanceToPrimaryInput(primaryInputAABB,feature.getGeometry())) // consider only polygons in range of the primary input
                    polygons.emplace_back(optId.value(),fileRef.value(),std::move(feature.getGeometry())); // create settlement wrapper containing its unique id and geometry
            }
        }
        return polygons;
    }

    void run() override{
        auto graph = fishnet::graph::GraphFactory::UndirectedGraph<SettlementPolygon<P>>(
            MemgraphAdjacency<SettlementPolygon<P>>(MemgraphClient(MemgraphConnection::create(config.params,workflowID).value_or_throw()))
        );   
        std::vector<SettlementPolygon<P>> polygons = readInput(graph);
        double maxEdgeDistanceVar = config.maxEdgeDistance;
        auto boundingBoxPolygonWrapper = [maxEdgeDistanceVar,this](const SettlementPolygon<P> & settPolygon ){
            /* Create scaled aaBB containing at least all points reachable from the polygon within the maximum edge distance*/
            auto aaBB = fishnet::geometry::Rectangle<fishnet::math::DEFAULT_NUMERIC>(settPolygon);
            double distanceMetersTopLeftBotLeft = distanceFunction({aaBB.left(),aaBB.top()},{aaBB.left(),aaBB.bottom()});
            double scale = (maxEdgeDistanceVar / distanceMetersTopLeftBotLeft) +1;
            return fishnet::geometry::BoundingBoxPolygon(settPolygon,aaBB.scale(scale));
        };
        fishnet::util::AllOfPredicate<P,P> neighbouringPredicate;
        /* add all neighbouring predicates to composite predicate */
        neighbouringPredicate.add(DistanceBiPredicate(distanceFunction,config.maxEdgeDistance));
        std::ranges::for_each(config.initNeighbouringPredicates<P>(),[&neighbouringPredicate](const auto & predicate){neighbouringPredicate.add(predicate);});
        auto shortCircuitPredicate = [neighbouringPredicate= std::move(neighbouringPredicate)](const fishnet::geometry::BoundingBoxPolygon<SettlementPolygon<P>> & lhs, const fishnet::geometry::BoundingBoxPolygon<SettlementPolygon<P>> & rhs){
            return lhs.getBoundingBox().overlap(rhs.getBoundingBox()) && neighbouringPredicate(lhs.getPolygon(),rhs.getPolygon());
        };
        auto result = fishnet::geometry::findNeighbouringPolygonsTemplate(polygons,shortCircuitPredicate,boundingBoxPolygonWrapper,config.maxNeighbours);    
        this->desc["Adjacencies"]=result.size();
        graph.addNodes(polygons);
        graph.addEdges(result);
    }
};